==> mark-search-0.2.10.tar.gz: mark-search-0.2.10/mark_search/query_executor.py <==
from .query_context import QueryContext
from .query_components import Range, Match, Terms, Term, And, Or
from .query_result import IdsQR, IdsScoresQR
from .combine_scores.wrapped import sorted_in1d, combine_must_should

from .query_validator import validateQuery

from .postings_prefetch import prefetchPostings

import numpy as np
import time

def filterHandler(value, qc):
    qc.setSetting('filter')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return And(qc)._filter(clauses)

def mustNotHandler(value, qc):
    qc.setSetting('filter')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return Or(qc)._filter(clauses)

def mustHandler(value, qc):
    qc.setSetting('score')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return And(qc)._score(clauses)

def shouldHandler(value, qc):
    qc.setSetting('score')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return Or(qc)._score(clauses)

def boolHandler(value, qc):

    # _filter = query_part_handlers['filter'](value.get('filter'), qc) if value.get('filter') else None
    # _must_not = query_part_handlers['must_not'](value.get('must_not'), qc) if value.get('must_not') else None
    # _must = query_part_handlers['must'](value.get('must'), qc) if value.get('must') else None
    # _should = query_part_handlers['should'](value.get('should'), qc) if value.get('should') else None

    has_filter = value.get('filter') is not None
    has_must_not = value.get('must_not') is not None
    has_must = value.get('must') is not None
    has_should = value.get('should') is not None

    has_any_filter = has_filter or has_must_not
    has_any_score = has_must or has_should

    has_filter_and_must_not = has_filter and has_must_not
    has_must_and_should = has_must and has_should

    if has_any_filter:

        combined_filter_ids = None
        combined_filter_invert = False
        if has_filter:
            if has_must_not:
                # if it has both, we can combine them to a single ids list
                _filter = query_part_handlers['filter'](value.get('filter'), qc)
                _must_not = query_part_handlers['must_not'](value.get('must_not'), qc)

                # using sorted_in1d with invert=True returns the ids in the first arg that are not in the second
                combined_filter_ids = _filter.getIds()[sorted_in1d(_filter.getIds(), _must_not.getIds(), invert=True, assume_sorted=True)]
                combined_filter_invert = False
            else:
                _filter = query_part_handlers['filter'](value.get('filter'), qc)
                combined_filter_ids = _filter.getIds()
                combined_filter_invert = False
        else:
            combined_filter_ids = query_part_handlers['must_not'](value.get('must_not'), qc).getIds()
            combined_filter_invert = True

        if has_must_and_should:
            _should = query_part_handlers['should'](value.get('should'), qc)
            _must = query_part_handlers['must'](value.get('must'), qc)

            filtered_must_mask = sorted_in1d(_must.getIds(), combined_filter_ids, invert=combined_filter_invert)
            _filtered_must_qr = IdsScoresQR(_must.getIds()[filtered_must_mask], _must.getScores()[filtered_must_mask])
            fm_s_ids, fm_s_scores = combine_must_should(_filtered_must_qr.getIds(), _filtered_must_qr.getScores(), _should.getIds(), _should.getScores(), assume_sorted=True)
            return IdsScoresQR(fm_s_ids, fm_s_scores)
        elif has_must:
            _must = query_part_handlers['must'](value.get('must'), qc)

            filtered_must_mask = sorted_in1d(_must.getIds(), combined_filter_ids, invert=combined_filter_invert, assume_sorted=True)
            _filtered_must_qr = IdsScoresQR(_must.getIds()[filtered_must_mask], _must.getScores()[filtered_must_mask])
            return _filtered_must_qr
        elif has_should:
            _should = query_part_handlers['should'](value.get('should'), qc)

            filtered_should_mask = sorted_in1d(_should.getIds(), combined_filter_ids, invert=combined_filter_invert, assume_sorted=True)
            _filtered_should_qr = IdsScoresQR(_should.getIds()[filtered_should_mask], _should.getScores()[filtered_should_mask])
            return _filtered_should_qr
        else:
            return IdsQR().fromIds(combined_filter_ids)
          
    else:
        # there are not filter / must_not
        # if there is must and should we combine, otherwise, return the one present
        if has_must_and_should:
            # combine must and should
            _should = query_part_handlers['should'](value.get('should'), qc)
            _must = query_part_handlers['must'](value.get('must'), qc)
            fm_s_ids, fm_s_scores = combine_must_should(_must.getIds(), _must.getScores(), _should.getIds(), _should.getScores(), assume_sorted=True)
            return IdsScoresQR(fm_s_ids, fm_s_scores)
        elif has_must:
            _must = query_part_handlers['must'](value.get('must'), qc)
            return _must
        else:
            _should = query_part_handlers['should'](value.get('should'), qc)
            return _should
    

def termHandler(value, qc):
    field = [key for key in value][0]
    o = value[field]
    term = o.get('value')
    boost = o.get('boost')
    
    if qc.setting == 'filter':
        return Term(qc)._filter(field, term)
    elif qc.setting == 'score':
        return Term(qc)._score(field, term, score=boost)
    else:
        assert False

def termsHandler(value, qc):
    field = [key for key in value][0]
    terms = value[field]
    
    if qc.setting == 'filter':
        return Terms(qc)._filter(field, terms)
    elif qc.setting == 'score':
        return Terms(qc)._score(field, terms)
    else:
        assert False

def matchHandler(value, qc):
    field = [key for key in value][0]
    terms = value[field]
    
    if qc.setting == 'filter':
        return Match(qc)._filter(field, terms)
    elif qc.setting == 'score':
        return Match(qc)._score(field, terms)
    else:
        assert False

def rangeHandler(value, qc):
    field = [key for key in value][0]
    bounds = value[field]

    if qc.setting == 'filter':
        return Range(qc)._filter(field, bounds=bounds)
    elif qc.setting == 'score':
        return Range(qc)._score(field, bounds=bounds)
    else:
        assert False

def fieldBoostHandler(value, qc):
    field = value.get('field')
    boost = value.get('boost', 1)

    if qc.setting == 'filter':
        assert False # this is a scoring only operation
    elif qc.setting == 'score':
        a = qc.index.getPostings(field, None)
        if boost != 1:
            # TODO: maybe in index.getPostings, we make a copy for the query
            b = IdsScoresQR(a.ids, np.array(a.scores, dtype=np.float32) * boost)
            return b
        else:
            return a
    else:
        assert False

def scoredTermHandler(value, qc):
    field = value['field']
    val = value['value']

    return  qc.index.getPostings(field, val)

def interleaveHandler(value, qc):
    queries = value['queries']
    n_from_each = value.get('n_from_each', 100)

    sub_query_results = []

    for query_name in queries:
        results = doQuery(queries[query_name],
            index=qc.index,
            limit=n_from_each,
            offset=0,
            min_score=0,
            return_scores=True,
            return_doc_ids=True,
            return_documents=False,
            prefetch=True
        )
        sub_query_results.append((results.get('docs', []), results.get('scores', [])))
    
    ids = []
    scores = []
    top_score = n_from_each * len(sub_query_results)

    for i in range(n_from_each):
        for j in range(len(sub_query_results)):
            if i < len(sub_query_results[j][0]):
                ids.append(sub_query_results[j][0][i])
                scores.append(top_score - i*j - j)
    
    return IdsScoresQR(np.array(ids, dtype=np.int32), np.array(scores, dtype=np.float32))


    #     ids = np.array(results.get('ids', []), dtype=np.int32)
    #     scores = np.array(results.get('scores', []), dtype=np.float32)

    #     sort_mask = np.argsort(ids)

    #     sub_query_results.append(IdsScoresQR(np.array(ids[sort_mask]), np.array(scores[sort_mask])))
    
    # return Interleave(qc)._score(sub_query_results)

query_part_handlers = {
    'bool': boolHandler,
    'filter': filterHandler,
    'must': mustHandler,
    'should': shouldHandler,
    'must_not': mustNotHandler,
    'term': termHandler,
    'terms': termsHandler,
    'match': matchHandler,
    'range': rangeHandler,
    'fieldBoost': fieldBoostHandler,
    'field_boost': fieldBoostHandler,
    'scored_term': scoredTermHandler,
    'interleave': interleaveHandler
}

def doQuery(query,
        index=None,
        limit=10,
        offset=0,
        min_score=0,
        return_scores=True,
        return_doc_ids=True,
        return_documents=False,
        prefetch=True):
    start_time = time.time()

    qc = QueryContext(index)
    qc.setSetting('score')

    validateQuery(query, index.mapping)

    validation_done_time = time.time()

    has_interleave = query.get('interleave') is not None

    if prefetch and not has_interleave:
        prefetchPostings(query, index)

    prefetch_done_time = time.time()

    entrypoint = [key for key in query][0]
    ids, scores = query_part_handlers[entrypoint](query[entrypoint], qc).getTopN(limit + offset)

    query_done_time = time.time()

    start = min(offset, ids.size)
    end = min(offset + limit, ids.size)

    doc_ids = [int(_id) for _id in ids[start:end]]

    response = {}
    if return_doc_ids:
        response['docs'] = doc_ids
    if return_scores:
        response['scores'] = [float(score) for score in scores[start:end]]
    if return_documents:
        response['hits'] = index.getDocuments(doc_ids) #[index.getDocument(i) for i in doc_ids]

    response_done_time = time.time()

    response['validation_took'] = validation_done_time - start_time
    response['prefetch_took'] = prefetch_done_time - validation_done_time
    response['query_took'] = query_done_time - prefetch_done_time
    response['response_took_total'] = response_done_time - start_time

    return response

==> mark-search-0.2.11.tar.gz: mark-search-0.2.11/mark_search/query_executor.py <==
from .query_context import QueryContext
from .query_components import Range, Match, Terms, Term, And, Or
from .query_result import IdsQR, IdsScoresQR
from .combine_scores.wrapped import sorted_in1d, combine_must_should

from .query_validator import validateQuery

from .postings_prefetch import prefetchPostings

import numpy as np
import time

def filterHandler(value, qc):
    qc.setSetting('filter')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return And(qc)._filter(clauses)

def mustNotHandler(value, qc):
    qc.setSetting('filter')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return Or(qc)._filter(clauses)

def mustHandler(value, qc):
    qc.setSetting('score')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return And(qc)._score(clauses)

def shouldHandler(value, qc):
    qc.setSetting('score')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return Or(qc)._score(clauses)

def boolHandler(value, qc):

    # _filter = query_part_handlers['filter'](value.get('filter'), qc) if value.get('filter') else None
    # _must_not = query_part_handlers['must_not'](value.get('must_not'), qc) if value.get('must_not') else None
    # _must = query_part_handlers['must'](value.get('must'), qc) if value.get('must') else None
    # _should = query_part_handlers['should'](value.get('should'), qc) if value.get('should') else None

    has_filter = value.get('filter') is not None
    has_must_not = value.get('must_not') is not None
    has_must = value.get('must') is not None
    has_should = value.get('should') is not None

    has_any_filter = has_filter or has_must_not
    has_any_score = has_must or has_should

    has_filter_and_must_not = has_filter and has_must_not
    has_must_and_should = has_must and has_should

    if has_any_filter:

        combined_filter_ids = None
        combined_filter_invert = False
        if has_filter:
            if has_must_not:
                # if it has both, we can combine them to a single ids list
                _filter = query_part_handlers['filter'](value.get('filter'), qc)
                _must_not = query_part_handlers['must_not'](value.get('must_not'), qc)

                # using sorted_in1d with invert=True returns the ids in the first arg that are not in the second
                combined_filter_ids = _filter.getIds()[sorted_in1d(_filter.getIds(), _must_not.getIds(), invert=True, assume_sorted=True)]
                combined_filter_invert = False
            else:
                _filter = query_part_handlers['filter'](value.get('filter'), qc)
                combined_filter_ids = _filter.getIds()
                combined_filter_invert = False
        else:
            combined_filter_ids = query_part_handlers['must_not'](value.get('must_not'), qc).getIds()
            combined_filter_invert = True

        if has_must_and_should:
            _should = query_part_handlers['should'](value.get('should'), qc)
            _must = query_part_handlers['must'](value.get('must'), qc)

            filtered_must_mask = sorted_in1d(_must.getIds(), combined_filter_ids, invert=combined_filter_invert)
            _filtered_must_qr = IdsScoresQR(_must.getIds()[filtered_must_mask], _must.getScores()[filtered_must_mask])
            fm_s_ids, fm_s_scores = combine_must_should(_filtered_must_qr.getIds(), _filtered_must_qr.getScores(), _should.getIds(), _should.getScores(), assume_sorted=True)
            return IdsScoresQR(fm_s_ids, fm_s_scores)
        elif has_must:
            _must = query_part_handlers['must'](value.get('must'), qc)

            filtered_must_mask = sorted_in1d(_must.getIds(), combined_filter_ids, invert=combined_filter_invert, assume_sorted=True)
            _filtered_must_qr = IdsScoresQR(_must.getIds()[filtered_must_mask], _must.getScores()[filtered_must_mask])
            return _filtered_must_qr
        elif has_should:
            _should = query_part_handlers['should'](value.get('should'), qc)

            filtered_should_mask = sorted_in1d(_should.getIds(), combined_filter_ids, invert=combined_filter_invert, assume_sorted=True)
            _filtered_should_qr = IdsScoresQR(_should.getIds()[filtered_should_mask], _should.getScores()[filtered_should_mask])
            return _filtered_should_qr
        else:
            return IdsQR().fromIds(combined_filter_ids)
          
    else:
        # there are not filter / must_not
        # if there is must and should we combine, otherwise, return the one present
        if has_must_and_should:
            # combine must and should
            _should = query_part_handlers['should'](value.get('should'), qc)
            _must = query_part_handlers['must'](value.get('must'), qc)
            fm_s_ids, fm_s_scores = combine_must_should(_must.getIds(), _must.getScores(), _should.getIds(), _should.getScores(), assume_sorted=True)
            return IdsScoresQR(fm_s_ids, fm_s_scores)
        elif has_must:
            _must = query_part_handlers['must'](value.get('must'), qc)
            return _must
        else:
            _should = query_part_handlers['should'](value.get('should'), qc)
            return _should
    

def termHandler(value, qc):
    field = [key for key in value][0]
    o = value[field]
    term = o.get('value')
    boost = o.get('boost')
    
    if qc.setting == 'filter':
        return Term(qc)._filter(field, term)
    elif qc.setting == 'score':
        return Term(qc)._score(field, term, score=boost)
    else:
        assert False

def termsHandler(value, qc):
    field = [key for key in value][0]
    terms = value[field]
    
    if qc.setting == 'filter':
        return Terms(qc)._filter(field, terms)
    elif qc.setting == 'score':
        return Terms(qc)._score(field, terms)
    else:
        assert False

def matchHandler(value, qc):
    field = [key for key in value][0]
    terms = value[field]
    
    if qc.setting == 'filter':
        return Match(qc)._filter(field, terms)
    elif qc.setting == 'score':
        return Match(qc)._score(field, terms)
    else:
        assert False

def rangeHandler(value, qc):
    field = [key for key in value][0]
    bounds = value[field]

    if qc.setting == 'filter':
        return Range(qc)._filter(field, bounds=bounds)
    elif qc.setting == 'score':
        return Range(qc)._score(field, bounds=bounds)
    else:
        assert False

def fieldBoostHandler(value, qc):
    field = value.get('field')
    boost = value.get('boost', 1)

    if qc.setting == 'filter':
        assert False # this is a scoring only operation
    elif qc.setting == 'score':
        a = qc.index.getPostings(field, None)
        if boost != 1:
            # TODO: maybe in index.getPostings, we make a copy for the query
            b = IdsScoresQR(a.ids, np.array(a.scores, dtype=np.float32) * boost)
            return b
        else:
            return a
    else:
        assert False

def scoredTermHandler(value, qc):
    field = value['field']
    val = value['value']

    return  qc.index.getPostings(field, val)

def interleaveHandler(value, qc):
    queries = value['queries']
    n_from_each = value.get('n_from_each', 100)

    sub_query_results = []

    for query_name in queries:
        results = doQuery(queries[query_name],
            index=qc.index,
            limit=n_from_each,
            offset=0,
            min_score=0,
            return_scores=True,
            return_doc_ids=True,
            return_documents=False,
            prefetch=True
        )
        sub_query_results.append((results.get('docs', []), results.get('scores', [])))
    
    ids = []
    scores = []
    curr_score = n_from_each * len(sub_query_results)

    for i in range(n_from_each):
        for j in range(len(sub_query_results)):
            if i < len(sub_query_results[j][0]):
                ids.append(sub_query_results[j][0][i])
                scores.append(curr_score)
                curr_score -= 1
    
    return IdsScoresQR(np.array(ids, dtype=np.int32), np.array(scores, dtype=np.float32))


    #     ids = np.array(results.get('ids', []), dtype=np.int32)
    #     scores = np.array(results.get('scores', []), dtype=np.float32)

    #     sort_mask = np.argsort(ids)

    #     sub_query_results.append(IdsScoresQR(np.array(ids[sort_mask]), np.array(scores[sort_mask])))
    
    # return Interleave(qc)._score(sub_query_results)

query_part_handlers = {
    'bool': boolHandler,
    'filter': filterHandler,
    'must': mustHandler,
    'should': shouldHandler,
    'must_not': mustNotHandler,
    'term': termHandler,
    'terms': termsHandler,
    'match': matchHandler,
    'range': rangeHandler,
    'fieldBoost': fieldBoostHandler,
    'field_boost': fieldBoostHandler,
    'scored_term': scoredTermHandler,
    'interleave': interleaveHandler
}

def doQuery(query,
        index=None,
        limit=10,
        offset=0,
        min_score=0,
        return_scores=True,
        return_doc_ids=True,
        return_documents=False,
        prefetch=True):
    start_time = time.time()

    qc = QueryContext(index)
    qc.setSetting('score')

    validateQuery(query, index.mapping)

    validation_done_time = time.time()

    has_interleave = query.get('interleave') is not None

    if prefetch and not has_interleave:
        prefetchPostings(query, index)

    prefetch_done_time = time.time()

    entrypoint = [key for key in query][0]
    ids, scores = query_part_handlers[entrypoint](query[entrypoint], qc).getTopN(limit + offset)

    query_done_time = time.time()

    start = min(offset, ids.size)
    end = min(offset + limit, ids.size)

    doc_ids = [int(_id) for _id in ids[start:end]]

    response = {}
    if return_doc_ids:
        response['docs'] = doc_ids
    if return_scores:
        response['scores'] = [float(score) for score in scores[start:end]]
    if return_documents:
        response['hits'] = index.getDocuments(doc_ids) #[index.getDocument(i) for i in doc_ids]

    response_done_time = time.time()

    response['validation_took'] = validation_done_time - start_time
    response['prefetch_took'] = prefetch_done_time - validation_done_time
    response['query_took'] = query_done_time - prefetch_done_time
    response['response_took_total'] = response_done_time - start_time

    return response
